var crypto = require('crypto');
var exec = require('child_process').exec;
var fs = require('fs');
var path = require("path");
var request = require('request');
var url = require('url');
var zlib = require('zlib');

const args = require('yargs')
    .options({
        'data-folder': {
            default: '/local/',
            type: 'string',
        },
        'job-folder': {
            default: '/root/start/',
            type: 'string',
        },
        'proxy': {
            default: true,
            type: 'boolean',
        },
        'shared-folder': {
            default: '/root/shared/',
            type: 'string',
        },
    })
    .argv
;

process.env.NODE_TLS_REJECT_UNAUTHORIZED = "0";

if (args.proxy) {
    var proxy = require('./proxy.js');
}

var GR = { // State
    activeState: false,
    fileChunkSize: 1024 * 1024, // file chunk size when uploading files
    key: undefined, // job instance key
    server: 'https://app-api.charityengine.services/remotejobs/v2/',
    sharedFolder: 'shared/', // Shared folder location
    dataFolder: 'local/', // Data directory shared with docker containers
    jobFolder: 'start/', // Directory for docker jobs to start
    states: { // list of states and their statuses
        default: {
            commandLine: undefined, // command line to execute
            inputFiles: {}, // list of input files
            status: 'initial', // One of initial, downloading, downloaded, running, completed, failed
            startTime: undefined, // Start time in unixtime
            endTime: undefined, // End time in unixtime
        },
    },
    timeLimit: 0,
    startTime: false, // Script start time
};

GR.startTime = timestamp();

GR.sharedFolder = args.sharedFolder;
GR.dataFolder = args.dataFolder;
GR.jobFolder = args.jobFolder;

GR.canExecuteDownload = true; // Marker to only allow one download per file

if (fs.existsSync(GR.dataFolder + 'input') === false) {
    fs.mkdirSync(GR.dataFolder + 'input');
}

if (fs.existsSync(GR.dataFolder + 'output') === false) {
    fs.mkdirSync(GR.dataFolder + 'output');
}

if (fs.existsSync(GR.jobFolder) === false) {
    fs.mkdirSync(GR.jobFolder);
}

// Read startup state
if (fs.existsSync(GR.sharedFolder + 'state.default.json') === false) {
    shutdownVm(false, 'Missing startup state');
}
try {
    // @todo encryption
    GR.states.default = JSON.parse(
        fs.readFileSync(GR.sharedFolder + 'state.default.json')
    );
    GR.states.default.status = 'initial';
    GR.timeLimit = GR.states.default.timeLimit;
    GR.key = GR.states.default.key;
} catch (e) {
    shutdownVm(false, 'Failed to parse startup state');
}

// Set up timer
// Validate timer; needs to be specified in seconds in 1h increments
if ((typeof GR.timeLimit === 'undefined') || (GR.timeLimit <= 0) || (GR.timeLimit % 3600 !== 0)) {
    shutdownVm(false, 'Incorrect timer');
}

// Shut down VM after reserved time runs out
setTimeout(function() {
    if (GR.activeState) {
        // Send message to server about termination
        GR.states[GR.activeState].status = 'aborted';

        // Update the state remotely
        updateRemoteState(GR.activeState)
            .then(function() {
                shutdownVm(true, 'Reserved time ran out');
            })
        ;
    }

    // In case the server state update fails or no state is active, shut down anyway after 10s
    setTimeout(function() {
        shutdownVm(true, 'Reserved time ran out, could not update server state');
    }, 10000);
}, GR.timeLimit * 1000);

GR.pollFrequency = 60000;
setTimeout(function() {
    runNextState();

    pollServer();
}, 5000);

function runNextState() {
    console.log('runNextState()');
    return new Promise(function(resolve, reject) {
        let stateName = chooseNextState();
        if (stateName === false) {
            // No pending states, pollServer() will eventually get and execute a new one.
            console.log('no inactive states');
            if (GR.states.default.shutdownWhenDone) {
                shutdownVm(true, 'All jobs completed');
            }
            return resolve(false);
        }

        GR.activeState = stateName;

        runState(stateName)
            .then(function(result) {
                runNextState();
                //resolve(result);
            }
        );
    });
}

function chooseNextState() {
    if (GR.states.default.status === 'initial') {
        return 'default';
    }

    for (let name in GR.states) {
        if (GR.states[name].status === 'initial') {
            return name;
        }
    }

    return false;
}

/**
 * Runs a specific state
 * @param stateName Name of the state to execute. Initial state is "default"
 */
function runState(stateName) {
    console.log('runState(' + stateName + ')');
    return new Promise(function(resolve, reject) {
        GR.states[stateName].status = 'downloading';

        // Update the state remotely
        updateRemoteState(stateName);

        // Read and download input files
        let promises = [];
        if (typeof GR.states[stateName].inputFiles !== 'undefined') {
            for (let i in GR.states[stateName].inputFiles) {
                if (GR.states[stateName].inputFiles[i] === null) {
                    continue;
                }
                promises.push(downloadFile(stateName, i));
            }
        }
        Promise
            .all(promises)
            .then(function(result) {
                return downloadsCompleted(stateName, result);
            })
            .then(function(result) {
                GR.activeState = false;
                resolve(result);
            });
    });
}

function downloadFileExecute(fullUrl, stateName, i, filename, resolve, reject, fd) {
    if (GR.canExecuteDownload === false) {
        return setTimeout(function() {
            downloadFileExecute(fullUrl, stateName, i, filename, resolve, reject, fd);
        }, 1000);
    }

    GR.canExecuteDownload = false; // Mark as download in progress

    console.log(timestamp(), 'Downloading file', fullUrl, filename);

    request.get(fullUrl)
        .on('end', function () {
            if (GR.states[stateName].inputFiles[i].size) {
                let inputFilePath = GR.dataFolder + 'input/' + filename;
                let stats = fs.statSync(inputFilePath);
                if (stats.size !== GR.states[stateName].inputFiles[i].size) {
                    console.log('Downloaded file size mismatch: expected ' + GR.states[stateName].inputFiles[i].size + ', got ' + stats.size);
                    fs.unlinkSync(inputFilePath);

                    GR.canExecuteDownload = true;
                    GR.states[stateName].inputFiles[i].status = 'failed';
                    GR.states[stateName].inputFiles[i].error = 'size mismatch';

                    // Retry
                    console.log(timestamp(), 'Retrying download', fullUrl, filename);
                    downloadFileExecute(fullUrl, stateName, i, filename, resolve, reject, fd);

                    return;
                }
            }

            GR.states[stateName].inputFiles[i].filename = filename;
            GR.states[stateName].inputFiles[i].status = 'downloaded';

            if (GR.states[stateName].inputFiles[i].cache) {
                // File should be cached
                let hash = crypto.createHash('md5').update(fullUrl).digest('hex');
                fs.createReadStream(filename).pipe(fs.createWriteStream('/root/scratch/' + hash + '.input'));
            }

            GR.canExecuteDownload = true;

            console.log(timestamp(), 'Downloaded file', fullUrl, filename);

            resolve();
        })
        .on('error', function (err) {
            GR.canExecuteDownload = true;
            console.log(err);
            GR.states[stateName].inputFiles[i].status = 'failed';
            GR.states[stateName].inputFiles[i].error = err;

            // Retry
            console.log(timestamp(), 'Retrying download', fullUrl, filename);
            downloadFileExecute(fullUrl, stateName, i, filename, resolve, reject, fd);
        })
        .pipe(fs.createWriteStream('', {fd: fd}))
    ;
}

function downloadFile(stateName, i) {
    return new Promise(function(resolve, reject) {
        // @todo encryption
        let fullUrl = GR.states[stateName].inputFiles[i].url;
        let parsedUrl = url.parse(fullUrl, true);
        let filename = GR.states[stateName].inputFiles[i].filename;
        if (!filename) {
            filename = path.basename(parsedUrl.pathname);
        }
        if (!filename) {
            filename = i;
        }
        let inputFilePath = GR.dataFolder + 'input/' + filename;

        // If input file already exists, version it
        while (fs.existsSync(inputFilePath)) {
            console.log('file already exists', inputFilePath);

            let parsed = path.parse(filename);
            filename = parsed.name + '_new' + parsed.ext;
            inputFilePath = GR.dataFolder + 'input/' + filename;
        }
        fs.open(inputFilePath, 'wx', function(err, fd) {
            // If a file was created because of a race condition, retry
            if ((err) && (err.code === 'EEXIST')) {
                downloadFile(stateName, i).then(function() {
                    resolve();
                });
                return;
            }

            // If input file is cacheable, md5 the URL and check if it is already cached
            if (GR.states[stateName].inputFiles[i].cache) {
                let hash = crypto.createHash('md5').update(fullUrl).digest('hex');

                fs.stat('/root/scratch/' + hash + '.input', function(error, stat) {
                    if (error) { // Does not yet exist
                        return downloadFileExecute(fullUrl, stateName, i, filename, resolve, reject, fd);
                    }

                    fs.createReadStream('/root/scratch/' + hash + '.input')
                        .pipe(fs.createWriteStream('', {fd: fd}))
                        .on('end', function() {
                            resolve();
                        })
                    ;
                });

                return;
            }

            downloadFileExecute(fullUrl, stateName, i, filename, resolve, reject, fd);
        });
    });
}

function awaitStateExecution(stateName) {
    if ((GR.states[stateName].status === 'downloading') && (fs.existsSync(GR.sharedFolder + stateName + '.log'))) {
        // The job has started running
        GR.states[stateName].status = 'running';
        updateRemoteState(stateName);
        return;
    }

    setTimeout(awaitStateExecution, 1000, stateName);
}

/**
 * Callback when all file downloads complete
 * @param stateName Name of the state that is being executed
 * @param status[] Array of statuses of all downloaded input files
 */
function downloadsCompleted(stateName, status) {
    return new Promise(function(resolve, reject) {
        if (GR.states[stateName].commandLine) {
            // Execute specified command line
            let commandLine = GR.states[stateName].commandLine;

            // Replace command line parameters with actual file names
            if (typeof GR.states[stateName].inputFiles !== 'undefined') {
                for (let i in GR.states[stateName].inputFiles) {
                    if (GR.states[stateName].inputFiles[i] !== null) {
                        let inputFilePath = GR.dataFolder + 'input/' + GR.states[stateName].inputFiles[i].filename;
                        commandLine.replace('%FILE' + (i + 1) + '%', inputFilePath);
                    }
                }
            }
            if (typeof GR.states['default'].inputFiles !== 'undefined') {
                for (let i in GR.states['default'].inputFiles) {
                    if (GR.states['default'].inputFiles[i] !== null) {
                        let inputFilePath = GR.dataFolder + 'input/' + GR.states['default'].inputFiles[i].filename;
                        commandLine.replace('%STARTFILE' + (i + 1) + '%', inputFilePath);
                    }
                }
            }

            let options = {
                maxBuffer: 1024 * 1024 * 1024, // default limit is 200KB, setting to 1GB instead
            };

            // Save computation start time
            GR.states[stateName].startTime = new Date().getTime();

            // Run the script

            // If a docker job, write a start file for boinc_app to run
            if (
                (GR.states.default.app.substr(0, 7) === 'docker:') ||
                (GR.states.default.app.substr(0, 11) === 'docker-x86:') ||
                (GR.states.default.app.substr(0, 14) === 'docker-nvidia:')
            ) {
                let startFile = GR.jobFolder + stateName;
                try {
                    let dockerImage = GR.states.default.app.substr(GR.states.default.app.indexOf(':') + 1);
                    let dockerCommand = dockerImage + '\n' + commandLine;
                    fs.writeFileSync(startFile, dockerCommand);
                } catch (e) {
                    console.log('error: ' + e);
                    GR.states[stateName].status = 'failed';
                    return reject();
                }

                // Once the start file is created, wait for it to be run
                awaitStateExecution(stateName);

                // Monitor for job completion
                fs.watch(path.dirname(startFile), function(event, filename) {
                    if (event === 'rename' && filename === stateName) {
                        if (!fs.existsSync(startFile)) {
                            // The start file has been removed, so it is done
                            GR.states[stateName].endTime = new Date().getTime();
                            GR.states[stateName].status = 'completed';

                            createSnapshot(stateName).then(function() {
                                console.log('created snapshot', stateName);
                                if (GR.states[stateName].cleanupFiles) {
                                    fs.readdir(GR.dataFolder + 'input', (err, files) => {
                                        for (const file of files) {
                                            console.log('removing input', file);
                                            fs.unlinkSync(GR.dataFolder + 'input/' + file);
                                        }
                                        fs.readdir(GR.dataFolder + 'output', (err, files) => {
                                            for (const file of files) {
                                                console.log('removing output', file);
                                                fs.unlinkSync(GR.dataFolder + 'output/' + file);
                                            }
                                            return resolve();
                                        });
                                        return;
                                    });
                                    return; // Only resolve after output file cleanup
                                }
                                return resolve();
                            });
                        }
                    }
                });

                return;
            }

            exec(commandLine, options, function callback(error, stdout, stderr) {
                GR.states[stateName].endTime = new Date().getTime();

                if (error) {
                    console.log('error: ' + error);
                    GR.states[stateName].status = 'failed';
                    return reject();
                }

                GR.states[stateName].status = 'completed';
                createSnapshot(stateName).then(function() {
                    console.log('created snapshot', stateName);
                    if (GR.states[stateName].cleanupFiles) {
                        let logFile = GR.sharedFolder + stateName + '.log';
                        console.log('removing log', logFile);
                        fs.unlinkSync(logFile);
                        fs.readdir(GR.dataFolder + 'input', (err, files) => {
                            for (const file of files) {
                                console.log('removing input', file);
                                fs.unlinkSync(GR.dataFolder + 'input/' + file);
                            }
                            fs.readdir(GR.dataFolder + 'output', (err, files) => {
                                for (const file of files) {
                                    console.log('removing output', file);
                                    fs.unlinkSync(GR.dataFolder + 'output/' + file);
                                }
                                return resolve();
                            });
                            return;
                        });
                        return; // Only resolve after output file cleanup
                    }
                    return resolve();
                });
            });
        }

        // No command line
        return resolve();
    });
}

function updateRemoteState(stateName) {
    return new Promise(function(resolve, reject) {
        console.log('Updating remote state', stateName);

        let logData = null;
        if (fs.existsSync(logFile = GR.sharedFolder + stateName + '.log')) {
            try {
                let logFile = GR.sharedFolder + stateName + '.log';
                logData = fs.readFileSync(logFile, 'utf8');
                if (logData.length > 4096) {
                    let excessChars = logData.length - 4096;
                    logData = logData.substring(excessChars, 4096);
                }
            } catch (e) {
                console.log('error reading state log: ' + e);
            }
        }

        let payload = {
            key: GR.key,
            stateName: stateName,
            state: GR.states[stateName],
            log: logData,
        };
        request
            .post({
                url: GR.server + 'update-state',
                timeout: 45000,
                body: JSON.stringify(payload),
                gzip: true,
            }, function(err, response, body) {
                console.log('update-state response', body);
                if (err) {
                    console.log(timestamp(), 'Retrying state update', stateName);
                    updateRemoteState(stateName)
                        .then(function(url) {
                            resolve(url);
                        })
                    ;
                    return;
                }
                let output = JSON.parse(body);
                if (output.success === false) {
                    console.log(timestamp(), 'Retrying state update', stateName);
                    updateRemoteState(stateName)
                        .then(function(url) {
                            resolve(url);
                        })
                    ;
                    return;
                }
                if (output.exists === false) {
                    return resolve(false);
                }
                return resolve(output.url);
            })
        ;
    });
}

function createSnapshot(stateName) {
    return new Promise(function(resolve, reject) {
        // Create new snapshot
        GR.states[stateName].outputFiles = {};

        fs.readdir(GR.dataFolder + 'output', function(err, files) {
            // Iterate output directory
            let promises = [];
            for (let i in files) {
                let filename = files[i];
                // Upload each file
                promises.push(uploadFile(filename));
            }
            Promise
                .all(promises)
                .then(function(status) {
                    // Merge result array into a single object. Filenames are
                    // guaranteed to be unique, so no data is lost.
                    GR.states[stateName].outputFiles = status.reduce(function(acc, x) {
                        for (var key in x) {
                            acc[key] = x[key];
                        }
                        return acc;
                    }, {});
                    // All uploads are done
                    updateRemoteState(stateName)
                        .then(function(status) {
                            console.log('state updated', JSON.stringify(GR.states[stateName], null, 4));
                            return resolve();
                        })
                    ;
                })
            ;
        });
    });
}

function uploadFile(file) {
    return new Promise(function(resolve, reject) {
        console.log('Uploading', file);
        fs.stat(GR.dataFolder + 'output/' + file, function(error, stat) {
            if (error) {
                return reject(error);
            }

            if(stat.isFile() === false) { // Directory
                return resolve(true); // not throwing, simply ignoring
            }

            // Read file in chunks
            let size = stat.size;
            let processed = 0;
            let promises = [];
            let i = 0;
            while (processed < size) {
                let uploadSize = GR.fileChunkSize;
                if (size - processed < GR.fileChunkSize) {
                    uploadSize = size - processed;
                }
                promises.push(uploadFilePart(file, i, processed, uploadSize));
                processed += uploadSize;
                i++;
            }
            Promise
                .all(promises)
                .then(function(status) {
                    // minimize file part information
                    let fileparts = [];
                    for (let part in status) {
                        fileparts[status[part].partNumber] = status[part];
                        delete fileparts[status[part].partNumber].partNumber;
                    }
                    console.log('all parts uploaded', fileparts);
                    // All parts have been uploaded
                    let result = {};
                    result[file] = {
                        size: size,
                        parts: fileparts,
                    };
                    return resolve(result);
                })
            ;
        });
    });
}

function uploadFilePart(file, partNumber, from, size) {
    return new Promise(function(resolve, reject) {
        console.log('Uploading file part', file, from, size);

        const chunks = [];
        fs
            .createReadStream(GR.dataFolder + 'output/' + file, {
                start: from,
                end: from + size - 1,
                autoClose: true
            })
            .on('data', function(chunk) {
                chunks.push(chunk);
            })
            .on('end', function() {
                let data = Buffer.concat(chunks);
                let hash = crypto.createHash('md5').update(data).digest('hex');
                checkFilePartPresence(hash)
                    .then(function(url) {
                        if (url !== false) {
                            // File part exists on the server
                            console.log('checkFilePartPresence()', url);
                            return resolve({hash: hash, partNumber: partNumber, url: url});
                        }

                        // Upload file part
                        processUploadFilePart(hash, data)
                            .then(function(status) {
                                return resolve({hash: hash, partNumber: partNumber, url: status});
                            })
                            .catch(function(error) {
                                console.log('Failed processUploadFilePart()', error);
                                return reject(error);
                            })
                        ;
                    })
                    .catch(function(error) {
                        console.log('Failed checkFilePartPresence()', hash, error);
                        return reject(error);
                    })
                ;
            })
        ;
    });
}

function retryProcessUploadFilePart(hash, data, resolve) {
    // Retry until success after a delay
    setTimeout(function () {
        processUploadFilePart(hash, data)
            .then(function (status) {
                resolve(status);
            })
        ;
    }, 5000);
}

function processUploadFilePart(hash, data) {
    return new Promise(function(resolve, reject) {
        console.log('Uploading file part to the server', hash);

        // @todo figure out how to use compression, binary data is expensive when encoded in JSON

        let payload = {
            hash: hash,
            data: data.toString('base64'),
        };
        //zlib.gzip(JSON.stringify(payload), function (error, gzippedPayload) {
            request
                .post({
                    url: GR.server + 'file-part-upload',
                    timeout: 45000,
                    body: JSON.stringify(payload),
                    //headers: {'content-type': 'application/json', 'Content-Encoding': 'gzip'}
                }, function(err, response, body) {
                    if (err || body === '') {
                        console.log('processUploadFilePart() error', err);
                        retryProcessUploadFilePart(hash, data, resolve);
                        return; // Do not reject, will eventually complete
                    }

                    let output = {success: false};
                    try {
                        output = JSON.parse(body);
                    } catch (e) {
                        console.log('processUploadFilePart() body parsing failure', body);
                    }

                    if (output.success === false) {
                        console.log('processUploadFilePart() success=false', output);
                        retryProcessUploadFilePart(hash, data, resolve);
                        return; // Do not reject, will eventually complete
                    }

                    // Success
                    return resolve(output.url);
                })
            ;
        //});
    });
}

function retryCheckFilePartPresence(hash, resolve) {
// Retry until success after a delay
    setTimeout(function () {
        checkFilePartPresence(hash)
            .then(function (status) {
                resolve(status);
            })
        ;
    }, 5000);
}

function checkFilePartPresence(hash) {
    return new Promise(function(resolve, reject) {
        console.log('Validating file part existence on the server', hash);

        let payload = {
            hash: hash,
        };
        request
            .post({
                url: GR.server + 'file-part-presence',
                timeout: 45000,
                body: JSON.stringify(payload),
                gzip: true,
            }, function(err, response, body) {
                if (err) {
                    console.log('checkFilePartPresence() error', err);
                    retryCheckFilePartPresence(hash, resolve); // Do not reject, will eventually complete
                    return;
                }
                let output = JSON.parse(body);
                if (output.success === false) {
                    console.log('checkFilePartPresence() success=false', output);
                    retryCheckFilePartPresence(hash, resolve);
                    return; // Do not reject, will eventually complete
                }
                if (output.exists === false) {
                    // File part does not exist and needs to be uploaded
                    return resolve(false);
                }
                return resolve(output.url);
            })
        ;
    });
}

/**
 * Poll server for new input files, snapshot and termination requests
 */
function pollServer() {
    setTimeout(pollServer, GR.pollFrequency);

    return new Promise(function(resolve, reject) {
        let payload = {
            key: GR.key,
        };
        console.log('pollServer() payload', payload);
        request
            .post({
                url: GR.server + 'poll',
                timeout: 45000,
                body: JSON.stringify(payload),
                gzip: true,
            }, function(err, response, body) {
                if (err) {
                    console.log(err);

                    // Polling failed, so just run the next state
                    if (GR.activeState === false) {
                        runNextState();
                    }
                    return;
                }

                console.log('pollServer() response received');
                try {
                    body = JSON.parse(body);
                } catch (e) {
                    body = {};
                }
                if (body.transitionState === 'terminate') {
                    shutdownVm(true, 'Shutdown initiated by server');
                }
                if (typeof body.pollFrequency !== 'undefined') {
                    GR.pollFrequency = body.pollFrequency;
                }

                for (let stateName in body.states) {
                    if (typeof GR.states[stateName] === 'undefined') {
                        GR.states[stateName] = body.states[stateName];
                    }
                }

//                console.log(GR.states);

                if (GR.activeState === false) {
                    runNextState();
                }
            })
        ;
    });
}

function timestamp() {
    let now = new Date();
    return now.getFullYear() + '-' + (now.getMonth() + 1) + '-' + now.getDate() + ' ' + now.getHours() + ':' + now.getMinutes() + ':' + now.getSeconds();
}

function shutdownVm(success, message = undefined) {
    console.log(timestamp(), 'Shutting down VM');
    if (message) {
        console.log(message);
    }

    if (fs.existsSync('inhibit_shutdown')) {
        return process.exit();
    }

    try {
        fs.writeFile(GR.sharedFolder + 'completion_trigger_file', '', function(){});
        setTimeout(function() {
            return process.exit();
        }, 1000);
        return;
    } catch (e) {
        console.log('error:', e);
        return false;
    }
    return process.exit();
}
